#include "umath.h"
#include "matvecalloc.h"
#include "dynapro.h"

#define MAX_COLS MAXSEQ
#define MAX_ROWS MAXSEQ

#define INF_INT 2147480000

#define STOP_TRACEBACK -1

/* Sequence 1 runs down   matrix index by row index i.
   Sequence 2 runs across matrix index by row index j. */

/* OUT: maximum score in sum matrix */
double dyn_pr_n2
(
 /* IN:  similarity matrix [0..rows-1][0..cols-1] */
 double **s,
 /* IN:  gap-open pen. vector for seq. 1 [0..rows-1] */
 int *iOpnPen,
 /* IN:  gap-extend pen. vec. for seq. 1 [0..rows-1] */
 double *iExtPen,
 /* IN:  gap-open pen. vector for seq. 2 [0..cols-1] */
 int *jOpnPen,
 /* IN:  gap-extend pen. vec. for seq. 2 [0..cols-1] */
 double *jExtPen,
 /* IN:  residue number for seq. 1 [0..rows-1] */
 int *iresnum,
 /* IN:  residue number for seq. 2 [0..cols-1] */
 int *jresnum,
 /* IN:  length of sequence 1 (Runs down matrix, indexed by i) */
 int rows,
 /* IN:  length of sequence 2 (Runs across matrix, indexed by j)*/
 int cols,
 /* OUT: sum matrix [0..rows-1][0..cols-1] */
 double **SS,
 /* OUT: traceback matrix with i index [0..rows-1][0..cols-1] */
 double **TrRow,
 /* OUT: traceback matrix with j index [0..rows-1][0..cols-1] */
 double **TrCol,
 /* OUT: i index where overall max occurs 0..rows-1 */
 int *iOverallMax,
 /* OUT: j index where overall max occurs 0..cols-1 */
 int *jOverallMax) {
  double TrialScore, ColMax, RowMax;
  int i,j,i2,j2, jRowMax;
  int iGapLength, jGapLength;
  double iGapPenalty, jGapPenalty;
  double OverallMax = -INF_INT;
  int iColMax[MAX_COLS];
  
  for(i=0; i< rows; i++) {
    SS[i][0] = s[i][0];
    TrRow[i][0] = STOP_TRACEBACK;
    TrCol[i][0] = STOP_TRACEBACK;
  }
  for(j=0; j< cols; j++) {
    SS[0][j] = s[0][j];
    TrRow[0][j] = STOP_TRACEBACK;
    TrCol[0][j] = STOP_TRACEBACK;
    iColMax[j] = 0;
  }
  
  for(i=1; i< rows; i++) {

    jRowMax = 0;
    
    for(j=1; j< cols; j++) {
      
      /* You are at position i,j, consider three
         possible maximum values: */
      
      /* 1. The value at SS[i-1][j-1] with no gaps */

      /* 2. The maximum on column j-1 up to row i-2. */
      
      ColMax = -INF_INT;
      if (iresnum[i-1] != iresnum[i]) { /* No gaps inside residues! - Allen */
	for (i2=iColMax[j-1]; i2<i-1; i2++) {
	  if ((iresnum[i2] != iresnum[i2+1]) &&
	      ((SS[i-1][j-1] <= 0) || (s[i-1][j-1] < 0) || (i2 == 0) ||
	       (j == (cols-1)) || (TrCol[i2][j-1] == j-2) ||
	       (TrCol[i2][j-1] == STOP_TRACEBACK))) { /* XXX */
	    iGapLength = i - i2 - 1;
	    iGapPenalty = /* XXX */
	      MIN(iOpnPen[i],iOpnPen[i2]) +
	      ((iGapLength - 1) * MIN(iExtPen[i],iExtPen[i2]));
	    TrialScore = SS[i2][j-1] - iGapPenalty;
	    if (TrialScore >= ColMax) {
	      iColMax[j-1] = i2;
	      ColMax = TrialScore;
	    }
	  }
	}
      }
      
      /* 3. The maximum on row i-1 up to column j-2. */
      
      RowMax = -INF_INT;
      if (jresnum[j-1] != jresnum[j]) {
	for (j2=jRowMax; j2<j-1; j2++) {
	  if ((jresnum[j2] != jresnum[j2+1]) &&
	      ((SS[i-1][j-1] <= 0) || (s[i-1][j-1] < 0) || (j2 == 0) ||
	       (i == (rows-1)) || (TrRow[i-1][j2] == i-2) ||
	       (TrRow[i-1][j2] == STOP_TRACEBACK))) { /* XXX */
	    jGapLength = j - j2 - 1;
	    jGapPenalty = /* XXX */
	      MIN(jOpnPen[j],jOpnPen[j2]) +
	      ((jGapLength - 1) * MIN(jExtPen[j],jExtPen[j2]));
	    TrialScore = SS[i-1][j2] - jGapPenalty;
	    if (TrialScore >= RowMax) {
	      jRowMax = j2;
	      RowMax = TrialScore;
	    }
	  }
	}
      }
      
      if /* Case 1 */ 
	(SS[i-1][j-1] >= ColMax &&
	 SS[i-1][j-1] >= RowMax) {
	  TrRow[i][j] =  i-1;
	  TrCol[i][j] =  j-1;
	  SS[i][j] = s[i][j] + SS[i-1][j-1];
      } else if /* Case 2 */
	(/* ColMax >= SS[i-1][j-1] .. True anyway! */
	 ColMax > RowMax) {
	  TrRow[i][j] =  iColMax[j-1];
	  TrCol[i][j] =  j-1;
	  SS[i][j] = s[i][j] + ColMax;
      } else if (RowMax > ColMax) { /* Case 3 */
	/* RowMax >= SS[i-1][j-1]  ..True anyway! */
	TrRow[i][j] =  i-1;
	TrCol[i][j] =  jRowMax;
	SS[i][j] = s[i][j] + RowMax;
      } else { /* Tied ColMax, RowMax */
	if (rows > cols) {
	  TrRow[i][j] =  iColMax[j-1];
	  TrCol[i][j] =  j-1;
	  SS[i][j] = s[i][j] + ColMax;
	} else if (cols > rows) {
	  TrRow[i][j] =  i-1;
	  TrCol[i][j] =  jRowMax;
	  SS[i][j] = s[i][j] + RowMax;
	} else {
	  iGapLength = i - iColMax[j-1] - 1;
	  jGapLength = j - jRowMax - 1;
	  if (jGapLength >= iGapLength) {
	    TrRow[i][j] =  iColMax[j-1];
	    TrCol[i][j] =  j-1;
	    SS[i][j] = s[i][j] + ColMax;
	  } else {
	    TrRow[i][j] =  i-1;
	    TrCol[i][j] =  jRowMax;
	    SS[i][j] = s[i][j] + RowMax;
	  }
	}
      }
      
      if (((i == (rows-1)) || (j == (cols-1))) && /* global alignment */
	  (SS[i][j] >= OverallMax)) {
	OverallMax = SS[i][j];
	*iOverallMax = i;
	*jOverallMax = j;
      }
    } /* j-loop */
  } /* i-loop */
  return OverallMax;
}

/* OUT: number of equivalenced residues = n */
int TrMat2Equivalences 
(
 /* IN:  length of sequence 1 */
 int rows,
 /* IN:  length of sequence 2 */
 int cols,
 /* IN:  traceback matrix with row index [0..rows-1][0..cols-1] */
 double **TrRow,
 /* IN:  traceback matrix with col index [0..rows-1][0..cols-1] */
 double **TrCol,
 /* IN:  row index where overall max occurs 0..rows-1 */
 int iOverallMax,
 /* IN:  col index where overall max occurs 0..cols-1 */
 int jOverallMax,
 /* OUT: seq. 1 equivalenced residues, using fortran conv. [0..n-1] */
 int *equiv1,
 /* OUT: seq. 2 equivalenced residues, using fortran conv. [0..n-1] */
 int *equiv2,
 /* OUT: number of chain breaks */
 int *nbrks){
  int ReversedEquiv1[MAX_ROWS];
  int ReversedEquiv2[MAX_COLS];
  int n = 0; 
  int i = iOverallMax;
  int j = jOverallMax;
  int ii,jj,nn;
  do {
    ii = (int)TrRow[i][j];
    jj = (int)TrCol[i][j];

    ReversedEquiv1[n] = i;
    ReversedEquiv2[n] = j;
    n++;
    if (ii != STOP_TRACEBACK && 
	!(ii == i - 1 && jj == j - 1)) {
      *nbrks = *nbrks + 1;
    }
    i = ii; j = jj;
  } while (i != STOP_TRACEBACK);
  
  for(i=0,nn=n; nn>0;i++,nn--) {
    /* Plus 1 changes from C to fortran array conventions ! */
    equiv1[i] = ReversedEquiv1[nn-1] + 1;
    equiv2[i] = ReversedEquiv2[nn-1] + 1; 
  }

  return i;
}
